{ Simple puzzle written for Delphi.
  Move the squares around to put them back in order.

  RA  5/95

note: compile with Range Checking OFF!
}
unit Pzl;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Menus, StdCtrls,
  Config, About, ExtCtrls;

type
  { Holds info about a spot in the puzzle }
  APlace = record
    Pos: integer;
    Val: integer;
    btnPtr: TButton;
  end;
  { This array is sized at runtime according to the
    size of the puzzle desired - must have range checking off }
  APlaceArray = array[1..1] of APlace;

  TForm1 = class(TForm)
    MainMenu1: TMainMenu;
    Quit1: TMenuItem;
    Puzzle1: TMenuItem;
    Shuffle1: TMenuItem;
    About1: TMenuItem;
    N1: TMenuItem;
    Configure1: TMenuItem;
    Panel1: TPanel;
    Label1: TLabel;
    Panel2: TPanel;
    Label2: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Shuffle1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure Quit1Click(Sender: TObject);
    procedure Configure1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    pzlState: ^APlaceArray;  { ptr to dynamic array of APlace }
    shuffled: boolean;       { True if puzzle is shuffled }
    numRows, numCols: integer; { size of puzzle }
    numShuffles: integer;    { number of moves in shuffle }
    numMoves: integer;       { number of moves made so far }
    procedure makePuzzle;    { create the puzzle }
    function doClick(btnNum: integer): boolean;   { click a spot }
    procedure checkCompleted; { if completed, applause }
    procedure doSwitch(i: integer; j: integer); { move a square }
    procedure destroyPuzzle;  { free dynamic allocations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

{ create the default puzzle (4x4) at form creation }
procedure TForm1.FormCreate(Sender: TObject);
var
  i: integer;
begin
  numRows:= 4;  { default size }
  numCols:= 4;
  numShuffles:= 160; { default # of moves in a shuffle }
                     { (Default is 10*number of squares }
  numMoves:= 0;
  Label1.Caption:= inttostr(numMoves);
  makePuzzle;
end;

{ create a puzzle, size is given by numRows and numCols }
procedure TForm1.makePuzzle;
var
  panelW, panelH: integer;
  btnW, btnH: integer;
  btnL, btnT: integer;
  i: integer;
begin
  { This could be done muy betta.
  { Next iteration will allow dynamic re-sizing of the form
    and buttons, if and when I get around to it. }
  Form1.Height:= numRows*32 + 50;
  Form1.Width:= numCols*20 + 50;
  panelW:= Panel2.Width - 4;
  panelH:= Panel2.Height - 6;
  { allocate memory for the puzzle }
  GetMem(pzlState, numRows*numCols*sizeof(APlace));
  { calculate the size of each button, to fill avail. space }
  btnW:= panelW div numCols;
  btnH:= panelH div numRows;
  for i:= 1 to numRows*numCols do
  begin
    { calc placement of buttons }
    btnL:= 2 + ((i-1) mod numCols) * btnW;
    btnT:= 2 + ((i-1) div numCols) * btnH;
    { Create, place, and show the buttons }
    pzlState^[i].BtnPtr:= TButton.Create(Self);
    pzlState^[i].BtnPtr.Parent:= Self;
    pzlState^[i].BtnPtr.SetBounds(btnL, btnT, btnW, btnH);
    pzlState^[i].Pos:= i;
    pzlState^[i].Val:= i; { starting val is just the position }
    pzlState^[i].BtnPtr.Caption:= inttoStr(i);
    pzlState^[i].BtnPtr.onClick:= Button1Click;
  end;
  { the empty spot has a val of 0, and no button appears there }
  pzlState^[numRows*numCols].Val:= 0;
  pzlState^[numRows*numCols].BtnPtr.Visible:= FALSE;
  shuffled:= FALSE;
end;

{ when any button is clicked, this is executed }
procedure TForm1.Button1Click(Sender: TObject);
var
  i: integer;
begin
  { find which button was clicked }
  for i:= 1 to numRows*numCols do
    if Sender = pzlState^[i].btnPtr then
      doClick(i);  { move square if possible }
  if shuffled then
    checkCompleted; { check if puzzle is back together }
end;

function TForm1.doClick(btnNum: integer): boolean;
begin
  { move the button (this is accomplished by swapping 'Val')
    if the empty space is adjacent }
  doClick:= FALSE;
  { move up if possible }
  if (btnNum > numCols) and
     (pzlState^[btnNum-numCols].Val = 0) then
  begin
    doSwitch(btnNum, btnNum-numCols); { move the square up }
    doClick:= TRUE;                   { square was moved }
    exit;
  end;
  { move left if possible }
  if ((btnNum mod numCols) <> 1) and
      (pzlState^[btnNum-1].Val = 0) then
  begin
    doSwitch(btnNum, btnNum-1);  { move the square left }
    doClick:= TRUE;              { square was moved }
    exit;
  end;
  { move right if possible }
  if ((btnNum mod numCols) <> 0) and
      (pzlState^[btnNum+1].Val = 0) then
  begin
    doSwitch(btnNum, btnNum+1); { move the square right }
    doClick:= TRUE;             { square was moved }
    exit;
  end;
  { move down if possible }
  if (btnNum <= (numRows-1)*numCols) and
     (pzlState^[btnNum+numCols].Val = 0) then
  begin
    doSwitch(btnNum, btnNum+numCols); { move the square down }
    doClick:= TRUE;                   { square was moved }
    exit;
  end;
end;

{ move the clicked button to the empty space by swapping
  'Val' and Visible }
procedure TForm1.doSwitch(i: integer; j: integer);
var
  tmp: integer;
  isVisible: boolean;
begin
  tmp:= pzlState^[i].Val;
  pzlState^[i].Val:= 0;      { vacated space is empty }
  isVisible:= pzlState^[i].btnPtr.Visible;
  pzlState^[i].btnPtr.Visible:= FALSE;
  pzlState^[j].Val:= tmp;    { show square at new (formerly empty) spot }
  pzlState^[j].btnPtr.Visible:= isVisible;
  pzlState^[j].btnPtr.Caption:= inttoStr(tmp);
  inc(numMoves);
  Label1.Caption:= inttostr(numMoves);
end;

{ shuffle the puzzle }
procedure TForm1.Shuffle1Click(Sender: TObject);
var
  i: integer;
  pick: integer;
begin
  randomize;
  for i:= 1 to numRows*numCols do
    pzlState^[i].BtnPtr.Visible:= FALSE;
  i:= 0;
  while i < numShuffles do
  begin
    { pick a random square and try to move it.
      If it moves, inc counter }
    pick:= random(numRows*numCols) + 1;
    if pzlState^[pick].Val <> 0 then
      if doClick(pick) then
        inc(i);
  end;
  for i:= 1 to numRows*numCols do
    if pzlState^[i].Val <> 0 then
      pzlState^[i].BtnPtr.Visible:= TRUE;
  shuffled:= TRUE;
  numMoves:= 0;
  Label1.Caption:= inttostr(numMoves);
end;

{ display the amazing, colossal about box }
procedure TForm1.About1Click(Sender: TObject);
begin
  formAbout.ShowModal;
end;

{ What??? }
procedure TForm1.Quit1Click(Sender: TObject);
begin
  destroyPuzzle;
  Close;
end;

{ display a congratulations box if puzzle has been shuffled
  then put back together correctly }
procedure TForm1.checkCompleted;
var
  i: integer;
  ret: integer;
begin
  { puzzle is put back together if each square is at the
    original position (i.e., Val = square number) }
  for i:= 1 to numRows*numCols-1 do
    if pzlState^[i].Val <> i then
      exit;
  ret:= Application.MessageBox('Congratulations', 'You Did It', 0);
  shuffled:= FALSE;
end;

{ bring up the configuration form }
procedure TForm1.Configure1Click(Sender: TObject);
begin
  if formConfig.ShowModal > 1 then
  begin
    destroyPuzzle;  { free dynamically allocated memory }
    numRows:= formConfig.SpinEdit2.Value;
    numCols:= formConfig.SpinEdit3.Value;
    numShuffles:= formConfig.SpinEdit1.Value;
    makePuzzle;     { and create a new puzzle according
                      to the updated configuration }
  end;
end;

{ Free all dynamically allocated memory, in preparation
  for the creation of a new puzzle }
procedure TForm1.destroyPuzzle;
var
  i: integer;
begin
  for i:= 1 to numRows*numCols do
    pzlState^[i].BtnPtr.Free;
  FreeMem(pzlState, numRows*numCols*sizeof(APlace));
end;

end.
